// SPDX-License-Identifier: BSD-3-Clause

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <time.h>

#include "os_graph.h"
#include "os_threadpool.h"
#include "log/log.h"
#include "utils.h"

#define NUM_THREADS		4

static atomic_int sum;
static os_graph_t *graph;
static os_threadpool_t *tp;
/* TODO: Define graph synchronization mechanisms. */
static pthread_mutex_t graph_mutex;

/* TODO: Define graph task argument. */
typedef int *task_argument;

static void do_process_node(void *arg);

static void process_node(unsigned int idx)
{
	/* TODO: Implement thread-pool based processing of graph. */
	pthread_mutex_lock(&graph_mutex);
	if (graph->visited[idx] != PROCESSING) {
		pthread_mutex_unlock(&graph_mutex);
		return;
	}
	graph->visited[idx] = DONE;
	pthread_mutex_unlock(&graph_mutex);

	/* We don't care about the ordering of the additions, just make sure they are atomic */
	atomic_fetch_add_explicit(&sum, graph->nodes[idx]->info, memory_order_relaxed);

	for (unsigned int i = 0; i < graph->nodes[idx]->num_neighbours; i++) {
		unsigned int neighbour = graph->nodes[idx]->neighbours[i];

		pthread_mutex_lock(&graph_mutex);
		if (graph->visited[neighbour] == NOT_VISITED) {
			graph->visited[neighbour] = PROCESSING;
			pthread_mutex_unlock(&graph_mutex);
			unsigned int *neighbour_idx = malloc(sizeof(*neighbour_idx));

			DIE(neighbour_idx == NULL, "malloc");
			*neighbour_idx = neighbour;
			enqueue_task(tp, create_task(do_process_node, neighbour_idx, free));
		} else {
			pthread_mutex_unlock(&graph_mutex);
		}
	}
}

static void do_process_node(void *arg)
{
	int *idx = arg;

	process_node(*idx);
}

int main(int argc, char *argv[])
{
	FILE *input_file;

	if (argc != 2) {
		fprintf(stderr, "Usage: %s input_file\n", argv[0]);
		exit(EXIT_FAILURE);
	}

	input_file = fopen(argv[1], "r");
	DIE(input_file == NULL, "fopen");

	graph = create_graph_from_file(input_file);

	/* TODO: Initialize graph synchronization mechanisms. */
	DIE(pthread_mutex_init(&graph_mutex, NULL) != 0, "pthread_mutex_init");

	tp = create_threadpool(NUM_THREADS);
	graph->visited[0] = PROCESSING;
	process_node(0);
	wait_for_completion(tp);
	destroy_threadpool(tp);

	pthread_mutex_destroy(&graph_mutex);

	printf("%d", sum);

	return 0;
}
